import type {SchoolClassData, WorkGroupData} from './common';
import {randomGroupData, randomString} from './rand';

enum ClassType {
  'OFFER' = 'offer',
  'CLASS' = 'class',
}

export interface SearchOptions {
  schoolSelection: string;
  idiSchoolyear?: string;
  divisClassType?: ClassType;
  source?: string;
  activationState?: boolean;
  isLmsGroup?: boolean;
  isIservGroup?: boolean;
  quickSearch?: string;
}

const setText = (key: string, value: string): void => {
  cy.get(` input[name="${key}"]`).clear().type(value);
};

const setOption = (key: string, value: string): void => {
  cy.get(`input[name="${key}"]`)
    .closest('.uFormElement')
    .within(() => {
      cy.get('.uButton')
        .click();
      cy.get(`#uComboBox__option--${value}`)
        .click();
    });
};

const processSearchOptions = (key: string, value: boolean | string | ClassType): void => {
  if (typeof value === 'string') {
    if (Object.values<string>(ClassType).includes(value) || key === 'schoolSelection') {
      setOption(key, value);
    } else {
      setText(key, value);
    }
  }
  if (typeof value === 'boolean') {
    setOption(key, value ? 'enabled' : 'disabled');
  }
};

export const submitGroupSearch = (searchOptions: SearchOptions): void => {
  cy.url().should('eq', `${Cypress.config().baseUrl}/`);
  Object.entries(searchOptions).map(([key, value]) => processSearchOptions(key, value));
  cy.get('input[name="idiSchoolyear"]').type('---').type('{enter}');
  const interceptKey = `getGroups-${randomString(16)}`;
  cy.intercept('GET', '/ucsschool/bff-groups/v1/groups*').as(interceptKey);
  cy.contains('.searchForm__buttons .uButton', 'Suchen').click();
  cy.get('.uStandbyWrapper').should('be.visible');
  cy.wait(`@${interceptKey}`).its('response.statusCode').should('eq', 200);
  cy.get('.uStandbyWrapper').should('not.exist');
};

export const createWorkGroup = (school: string, workGroupData?: WorkGroupData): WorkGroupData => {
  let interceptKey = `listViewConf-${randomString(16)}`;
  cy.intercept('GET', '/ucsschool/bff-groups/v1/pageconf/listView').as(interceptKey);
  cy.visit('/');
  cy.wait(`@${interceptKey}`).its('response.statusCode').should('eq', 200);
  cy.get('.listView').should('be.visible');
  cy.get('.uStandbyWrapper').should('not.exist');
  interceptKey = `addViewConf-${randomString(16)}`;
  cy.intercept('GET', '/ucsschool/bff-groups/v1/pageconf/addView').as(interceptKey);
  cy.contains('.uGrid__header .uButton', 'Arbeitsgruppe hinzufügen')
    .click();
  cy.wait(`@${interceptKey}`).its('response.statusCode').should('eq', 200);
  cy.get('.editView').should('be.visible');
  cy.get('.uStandbyWrapper').should('not.exist');
  const workGroup = workGroupData ?? randomGroupData();
  cy.get('input[name="name"]').type(workGroup.name, {scrollBehavior: 'center'});
  cy.get('input[name="description"]').type(workGroup.description, {scrollBehavior: 'center'});
  cy.get('input[name="school"]')
    .closest('.uFormElement')
    .within(() => {
      cy.get('.uButton')
        .click();
      cy.get('.uComboBox__list__option')
        .should('have.length.above', 1)
        .get('div').contains(school)
        .click();
    });
  interceptKey = `createGroup-${randomString(16)}`;
  cy.intercept('POST', '/ucsschool/bff-groups/v1/workgroup').as(interceptKey);
  cy.contains('.listView__header__buttons .uButton', 'Arbeitsgruppe anlegen')
    .click();
  cy.wait(`@${interceptKey}`).its('response.statusCode').should('eq', 201);
  cy.url().should('eq', `${Cypress.config().baseUrl}/`);
  // the next two assertions make sure the tests work in firefox: Because
  //  firefox will cause the application to trigger a NetworkError if an
  //  on-going AJAX request is aborted. So we will wait for all AJAX
  //  requests to finish before considering this step to be finished.
  cy.get('.listView').should('be.visible');
  cy.get('.uStandbyWrapper').should('not.exist');
  return workGroup;
};

export const createMultipleWorkGroups = (school: string, workGroups: number | WorkGroupData[]): WorkGroupData[] => {
  let workGroupsList: WorkGroupData[] = [];
  if (typeof workGroups === 'number') {
    for (let i = 0; i < workGroups; i++) {
      workGroupsList.push(randomGroupData());
    }
  } else {
    workGroupsList = workGroups;
  }
  return workGroupsList.map(() => createWorkGroup(school));
};


const _findGroupRowIfExists = (workGroup: WorkGroupData): Cypress.Chainable<JQuery<HTMLElement>> => cy.get('.uGrid').then($grid => {
  const $node = $grid.find(`:contains(${workGroup.name})`);
  if ($node.length) {
    return $node.closest('tr');
  }
  return $node;
}).then($node => {
  if ($node.length > 0) {
    expect($node.find('td:nth-child(2)')).to.contain.text(workGroup.name);
  }
  return $node;
});

const _setPage = (page: number): Cypress.Chainable<boolean> => cy.get('.uGrid').then($grid => {
  if (page < 1) {
    return cy.wrap(false).then(() => false);
  }
  const $select = $grid.find('.uTable__pagination select[name="page"]');
  if ($select.length === 0) {
    const result = page === 1;
    return cy.wrap(result).then(() => result);
  }
  const $lastOption = $select.find('option:last-child');
  const lastPage = $lastOption.length ? parseInt($lastOption.val() as string, 10) : 1;
  if (page > lastPage) {
    return cy.wrap(false).then(() => false);
  }
  const newValue = String(page);
  if ($select.val() === newValue) {
    return cy.wrap(true).then(() => true);
  }
  if ($select.val() === null) {
    // The select is disabled. This happens if there is no data in the table.
    //  We will return true iff page #1 was requested (as that is the only
    //  available option in such a case)
    return cy.wrap(newValue === '1').then(result => result);
  }
  cy.wrap($select).select(String(page));
  // The next line is flaky: we are waiting a tiny bit to make sure the table
  //  content will be re-rendered before moving on. It would be much nicer to
  //  have a [data-page="X"] attribute on the table DOM to avoid having to
  //  wait for the table to reload like this.
  // eslint-disable-next-line cypress/no-unnecessary-waiting
  return cy.wait(100).then(() => true);
});

const _determineCurrentPage = (): Cypress.Chainable<number> => cy.get('.uGrid').then($grid => {
  const $select = $grid.find('.uTable__pagination select[name="page"]');
  return $select.length === 0 ? 1 : parseInt($select.val() as string, 10);
});

const _determineLastPage = (): Cypress.Chainable<number> => cy.get('.uGrid').then($grid => {
  const $option = $grid.find('.uTable__pagination select[name="page"] option:last-child');
  return $option.length === 0 ? 1 : parseInt($option.val() as string, 10);
});

const _findGroupRowByPaging = (workGroup: WorkGroupData, direction: 1 | -1): Cypress.Chainable<JQuery<HTMLElement>> => _findGroupRowIfExists(workGroup)
  .then($node => $node.length
    ? cy.wrap($node)
    : _determineCurrentPage().then(page => _setPage(page + direction))
      .then(foundNextPage => foundNextPage
        ? _findGroupRowByPaging(workGroup, direction)
        : _findGroupRowIfExists(workGroup)));

export const findGroupRow = (
  group: WorkGroupData | SchoolClassData,
  options?: {
    assertExists?: boolean;
    searchForward?: boolean;
  },
): Cypress.Chainable<JQuery<HTMLElement>> => (options?.searchForward
  ? _setPage(1)
  : _determineLastPage()
    .then(page => _setPage(page)))
  .then(success => {
    expect(success).to.be.equal(true);
    return options?.searchForward
      ? _findGroupRowByPaging(group, 1)
      : _findGroupRowByPaging(group, -1);
  })
  .then($row => {
    if (options?.assertExists) {
      expect($row).to.have.length.above(0, `Finding Group "${group.name}"`);

      // Align to 'end' of scrollable area since the table header can overlap the row with
      // `block: 'start' (which is the default)
      $row[0].scrollIntoView({block: 'end'});
    }
    return $row;
  });
